/**********************************************************************
    numero.cc    --  La clase "Numero" implementa un numero grande.
                    Esta clase guarda los numeros en un "ASCII retocado"
                    es decir, en vez de guardar las cifras en el mismo ASCII
                    guarda en bytes el valor de cada cifra.
  
    RealNumbers  1.0
    Copyright (C) 1997  Antonio Larrosa Jimenez

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
***********************************************************************/


#include "numero.h"
#include "newdebug.h"
#include <stdio.h>
#include <stdlib.h>
#include "messbox.h"
#include "numeroz.h"
#include "numfunc.h"

Numero::Numero()
{
Err=No_hay_error;
Base=BASE_N;
Tabla_multiplicar=NULL;
signo=positivo;
//maxPE=1;
//PE=(Cifra *)fmalloc(maxPE,IDfmPE);
maxPE=0;
PE=NULL;
// *PE=0;
nPE=0;
//maxPD=1; PD=(Cifra *)fmalloc(maxPD,IDfmPD);
maxPD=0;
PD=NULL;
// *PD=0;
nPD=0;
};

Numero::Numero(Numero &a)
{
Err=a.Err;
Base=a.Base;
Tabla_multiplicar=NULL;
signo=a.signo;
nPE=a.nPE;
nPD=a.nPD;
maxPE=a.maxPE;
maxPD=a.maxPD;
Cifra *psrc;
Cifra *ptgt;
ulong l;
if (maxPE==0)
   PE=NULL;
  else
   {
   PE=(Cifra *)fmalloc(maxPE,IDfmPE);
   psrc=a.PE;
   ptgt=PE;
   l=maxPE;
   while (l>0)
      {
      *ptgt=*psrc;
      ptgt++;
      psrc++;
      l--;
      };
   };
if (maxPD==0)
   PD=NULL;
  else
   {
   PD=(Cifra *)fmalloc(maxPD,IDfmPD);
   psrc=a.PD;
   ptgt=PD;
   l=maxPD;
   while (l>0)
      {
      *ptgt=*psrc;
      ptgt++;
      psrc++;
      l--;
      };
   };
};

Numero::Numero(NumeroZ &a)
{
Err=No_hay_error;
Base=a.Base;
Tabla_multiplicar=NULL;
signo=a.signo;
nPE=a.nPE;
nPD=0;
maxPE=a.maxPE;
maxPD=1;
Cifra *psrc;
Cifra *ptgt;
ulong l;
if (maxPE==0)
   PE=NULL;
  else
   {
   PE=(Cifra *)fmalloc(maxPE,IDfmPE);
   psrc=a.PE;
   ptgt=PE;
   l=maxPE;
   while (l>0)
      {
      *ptgt=*psrc;
      ptgt++;
      psrc++;
      l--;
      };
   };

PD=(Cifra *)fmalloc(maxPD,IDfmPD);
*PD=0;

};

Numero::Numero(long l)
{
Err=No_hay_error;
Base=BASE_N;
Tabla_multiplicar=NULL;
if ((0<=l)&&(l<Base))
   {
   signo=positivo;
   maxPE=1;
   PE=(Cifra *)fmalloc(maxPE,IDfmPE);
   *PE=l;
   nPE=1;
//   maxPD=1;
   maxPD=0;
//   PD=(Cifra *)fmalloc(maxPD,IDfmPD);
   PD=NULL;
//   *PD=0;
   nPD=0;
   return;
   };
char s[11];
sprintf(s,"%ld",l);
operator =(s);
};

Numero::Numero(ulong pe,ulong pd,bool inicializa)
{
Err=No_hay_error;
Base=BASE_N;
Tabla_multiplicar=NULL;
signo=positivo;
//maxPE=(pe==0)?(1):(pe); // Siempre debe tener una cifra reservada (por si acaso)
maxPE=pe;
nPE=pe;
if (maxPE>0) PE=(Cifra *)fmalloc(maxPE,IDfmPE);
   else PE=NULL;
// *PE=0;
//maxPD=(pd==0)?(1):(pd); // Siempre debe tener una cifra reservada (por si acaso)
maxPD=pd;
nPD=pd;
if (maxPD>0) PD=(Cifra *)fmalloc(maxPD,IDfmPD);
   else PD=NULL;
// *PD=0;
if (inicializa==SI)
   {
   Cifra *p=(Cifra *)PE;
   ulong l=maxPE;
   while (l>0)
       {
       *p=0;
       p++;
       l--;
       };
   l=maxPD;
   p=PD;
   while (l>0)
       {
       *p=0;
       p++;
       l--;
       };
   };
};

Numero::~Numero()
{
if (PE!=NULL) {ffree(PE);PE=NULL;};
if (PD!=NULL) {ffree(PD);PD=NULL;};
nPE=0;
nPD=0;
if (Tabla_multiplicar!=NULL)
    {
    ulong l=0;
    while (l<Base)
       {
       if (Tabla_multiplicar[l]!=NULL) delete Tabla_multiplicar[l];
       l++;
       };
    delete Tabla_multiplicar;
    };
};

Numero Numero::Equal(Cifra c)
{
Nuevo_tamano_PE(1);
nPE=1;
*(PE+maxPE-1)=c;
nPD=0;
return (*this);
};

Numero Numero::operator =(long l)
{
if (l==0)
    {
    nPE=0;
    nPD=0;
    return (*this);
    };
if ((0<l)&&(l<Base))
   {
   signo=positivo;
   Nuevo_tamano_PE(1);
   *(PE+maxPE-1)=l;
   nPE=1;
   nPD=0;
   return (*this); 
   };
char s[11];
sprintf(s,"%ld",l);
operator =(s);
return (*this);
};

Numero Numero::operator =(Numero a)
{
if (maxPE<a.nPE)
   {
   if (PE!=NULL) ffree(PE);
   maxPE=a.maxPE;
   PE=(Cifra *)fmalloc(maxPE,IDfmPE);
   };
nPE=a.nPE;
Base=a.Base;
Err=a.Err;

Cifra *psrc=a.PE+(a.maxPE-a.nPE);
Cifra *ptgt=PE+(maxPE-nPE);
ulong l=nPE;
while (l>0)
   {
   *ptgt=*psrc;
   ptgt++;
   psrc++;
   l--;
   };
if (maxPD<a.nPD)
   {
   if (PD!=NULL) ffree(PD);
   maxPD=a.maxPD;
   PD=(Cifra *)fmalloc(maxPD,IDfmPE);
   };
nPD=a.nPD;

//psrc=a.PD+(a.maxPD-a.nPD);
//ptgt=PD+(maxPD-nPD);
psrc=a.PD;
ptgt=PD;
l=nPD;
while (l>0)
   {
   *ptgt=*psrc;
   ptgt++;
   psrc++;
   l--;
   };
signo=a.signo;
return (*this);
};

Numero Numero::operator =(char *s)
{
if (Es_un_numero_correcto(s,Base,"-.")==NO) return (*this);
if (PE!=NULL) ffree(PE);
if (PD!=NULL) ffree(PD);
if (*s=='-')
   {
   signo=negativo;
   s++;
   }
  else
   if (*s=='+')
      {
      signo=positivo;
      s++;
      }
     else
      signo=positivo;

char *pc=(char *)s;
nPE=0;
nPD=0;
ulong *nP=&nPE;
ulong l;
while (*pc!=0)
   {
   if (*pc=='.')
      {
      nP=&nPD;
      }
     else
      (*nP)++;
   pc++;
   };
pc=(char *)s;
bool ok=SI;
if (nPE==0)
   {
//   maxPE=1;
//   PE=(Cifra *)fmalloc(maxPE,IDfmPE);
//   *PE=0;
   maxPE=0;
   PE=NULL;
   }
  else
   {
   maxPE=nPE;
   PE=(Cifra *)fmalloc(maxPE,IDfmPE);
   l=nPE;
   Cifra *pe=(Cifra *)PE;
   while (l>0)
      {
      *pe=Caracter_a_Cifra(*pc,ok);
      if (ok==NO) {nPE=0;return (*this);};
      pe++;
      pc++;
      l--;
      };
   };
if (nPD==0)
   {
//   maxPD=1;
//   PD=(Cifra *)fmalloc(maxPD,IDfmPD);
//   *PD=0;
   maxPD=0;
   PD=NULL;
   }
  else
   {
   maxPD=nPD;
   PD=(Cifra *)fmalloc(maxPD,IDfmPD);
   l=nPD;
   pc++; // para saltarse el punto decimal
   Cifra *pd=(Cifra *)PD;
   while (l>0)
      {
      *pd=Caracter_a_Cifra(*pc,ok);
      if (ok==NO) {nPD=0;return (*this);};
      pd++;
      pc++;
      l--;
      };
   };
Minimiza_PE();
Minimiza_PD();
return (*this);
};

ostream &operator<<(ostream& s,Numero arg)
{
  switch (arg.Err) 
     {
     case (Division_x_cero) : s << "Division por cero" ; break;
     case (Raiz_de_numero_negativo) : s << "El resultado es complejo" ; break;
     default : 
               char *pstr=(char *)fmalloc(arg.Longitud()+3,0);
               arg.Convierte_a_cadena(pstr);
               s << pstr;
               ffree(pstr);
               if (arg.Err==Cociente_periodico) s << "...";
   };
  return s;
};


void Numero::Convierte_a_cadena(char *s)
{
Minimiza_PE();
Minimiza_PD();
char *pc=(char *)s;
if (signo==negativo)
   {
   *pc='-';
   pc++;
   };
Cifra *p=(Cifra *)PE;
ulong l;
if (nPE==0)
   {
   *pc='0';
   pc++;
   }
  else   // nPE>0
   {
   l=nPE;
   p+=(maxPE-nPE);
   while (l>0)
      {
      *pc=Cifra_a_Caracter(*p);
      pc++;
      p++;
      l--;
      };
   };
if (nPD>0)
   {
   *pc='.';
   pc++;
   l=nPD;
   p=(Cifra *)PD;
   while (l>0)
      {
      *pc=Cifra_a_Caracter(*p);
      pc++;
      p++;
      l--;
      };
   };
*pc=0;
};

Cifra Numero::Caracter_a_Cifra(char c,bool &ok)
{
ok=SI;
if (((c)>='0')&&((c)<='9')) return (c)-'0';
if (((c)>='A')&&((c)<='Z')) return ((c)-'A')+10;
if (((c)>='a')&&((c)<='z')) return ((c)-'a')+10;
ok=NO;
return CIFRA_IS_NAN;
};

char Numero::Cifra_a_Caracter(Cifra c)
{
if (c<10) return '0'+(c);
if (c==CIFRA_IS_NAN) return '?';
return 'A'+(c-10);
};

Numero Numero::operator +(Numero &b)
{
if (signo!=b.signo) // Es una resta
   {
   Numero resultado;
   if (signo==positivo)  //  (+a)+(-b) = (+a)-(+b) = (a-b)
      {
      b.signo=positivo;
      resultado=(*this)-b;
      b.signo=negativo;
      }
     else                //  (-a)+(+b) = (+b)+(-a) = (b-a)
      {
      signo=positivo;
      resultado=b-(*this);
      signo=negativo;
      };
   return resultado;
   };

ulong maxnPE=MAXl(nPE,b.nPE);
ulong maxnPD=MAXl(nPD,b.nPD);
Numero c(maxnPE+1,maxnPD);
//Numero c;
//c.Nuevo_tamano_PE(maxnPE+1);
//c.Nuevo_tamano_PD(maxnPD);
c.nPD=maxnPD;
Cifra carry=0;
ulong l=maxnPD;
Cifra *pc=c.PD+(c.maxPD-1);
Cifra ca;
Cifra cb;
while (l>0)
    {
    ca=Cifra_en_posicion_PD(l);
    cb=b.Cifra_en_posicion_PD(l);
    *pc=(ca+cb+carry);
    if (*pc>=Base)
       {
       (*pc)-=Base;
       carry=1;
       }
      else
       carry=0;
    pc--;
    l--;
    };
l=0;
pc=c.PE+(c.maxPE-1);
while (l<maxnPE)
    {
    ca=Cifra_en_posicion_PE(l);
    cb=b.Cifra_en_posicion_PE(l);
    *pc=(ca+cb+carry);
    if (*pc>=Base)
       {
       (*pc)-=Base;
       carry=1;
       }
      else
       carry=0;
    pc--;
    l++;
    };
if (carry==0)
   {
   c.nPE=maxnPE;
   }
  else
   {
   c.nPE=maxnPE+1;
   *pc=carry;
   };
//c.Minimiza_PD();
if (signo==negativo) c.signo=negativo;
return c;
};

void Numero::Incrementa(Numero &b)
{
ulong maxnPE=MAXl(nPE,b.nPE);
Nuevo_tamano_PE(maxnPE+1);
Cifra carry=0;
ulong l=0;
Cifra *pc=PE+maxPE-1;
Cifra ca,cb;
while (l<maxnPE)
    {
    ca=Cifra_en_posicion_PE(l);
    cb=b.Cifra_en_posicion_PE(l);
    *pc=(ca+cb+carry);
    if (*pc>=Base)
       {
       (*pc)-=Base;
       carry=1;
       }
      else
       carry=0;
    pc--;
    l++;
    };
if (carry==0)
   {
   nPE=maxnPE;
   }
  else
   {
   nPE=maxnPE+1;
   *pc=carry;
   };
};

Cifra Numero::Cifra_en_posicion_PE(ulong l)
{
if (l>=nPE) return 0;
Cifra *p=PE;
p+=(maxPE-1)-l;
return *p;
};

Cifra Numero::Cifra_en_posicion_PD(ulong l)
{
if (l>nPD) return 0;
Cifra *p=PD;
p+=(l-1);
return *p;
};
void Numero::Minimiza_PE()
{
while ((nPE>0)&&(Cifra_en_posicion_PE(nPE-1)==0)) nPE--;
if ((nPE==0)&&(nPD==0)) signo=positivo;
};

void Numero::Minimiza_PD()
{
while ((nPD>0)&&(Cifra_en_posicion_PD(nPD)==0)) nPD--;
if ((nPE==0)&&(nPD==0)) signo=positivo;
};

Numero Numero::operator -(Numero &b)
{
if (signo!=b.signo) // Es una suma
   {
   Numero resultado;
   if (signo==positivo)  //  (+a)-(-b) = (+a)+(+b) = (a+b)
      {
      b.signo=positivo;
      resultado=(*this)+b;
      b.signo=negativo;
      }
     else                //  (-a)-(+b) = (-a)+(-b) = -(a+b)
      {
      signo=positivo;
      resultado=(*this)+b;
      signo=negativo;
      resultado.signo=negativo;
      };
   return resultado;
   };
ulong maxPD=MAXl(nPD,b.nPD);
ulong maxPE=MAXl(nPE,b.nPE);
Numero c(maxPE+1,maxPD);
if (signo==positivo)  //Los dos son positivos
   {
   if ((*this)<b)
      {
      c=b-(*this);
      c.signo=negativo;
      return c;
      };
   }
  else  // Los dos son negativos
   {
   if (b<(*this))
      {
      b.signo=positivo;
      signo=positivo;
      c=b-(*this);
      b.signo=negativo;
      signo=negativo;
      c.signo=negativo;
      return c;
      };
   };

Cifra carry=0;
ulong l=maxPD;
Cifra *pc=c.PD+maxPD-1;
Cifra ca;
Cifra cb;
while (l>0)
    {
    ca=Cifra_en_posicion_PD(l);
    cb=b.Cifra_en_posicion_PD(l)+carry;
    if (ca<cb)
       {
       *pc=(ca+Base)-cb;
       carry=1;
       }
      else
       {
       *pc=(ca-cb);
       carry=0;
       };
    pc--;
    l--;
    };
l=0;
pc=c.PE+maxPE;
while (l<maxPE)
    {
    ca=Cifra_en_posicion_PE(l);
    cb=b.Cifra_en_posicion_PE(l)+carry;
    if (ca<cb)
       {
       *pc=(ca+Base)-cb;
       carry=1;
       }
      else
       {
       *pc=(ca-cb);
       carry=0;
       };
    pc--;
    l++;
    };
*pc=carry;
c.Minimiza_PE();
c.Minimiza_PD();
return c;
};

Numero Numero::operator -(void)
{
Numero Resultado(*this);
Resultado.Cambia_signo();
return Resultado;
};

int Numero::operator <(Numero &b)
{
if (signo!=b.signo)
   {
   if (signo==negativo) return SI;
   return NO;
   };
// Tienen el mismo signo
bool si=SI;
bool no=NO;
if (signo==negativo) // si es negativo el mas grande es el de menos valor
   {
   si=NO;
   no=SI;
   };
if (nPE<b.nPE) return si;
if (nPE>b.nPE) return no;
ulong l=nPE;
Cifra ca,cb;
while (l>0)
    {
    ca=Cifra_en_posicion_PE(l-1);
    cb=b.Cifra_en_posicion_PE(l-1);
    if (ca<cb) return si;
      else if (ca>cb) return no;
    l--;
    };

ulong minPD=MINl(nPD,b.nPD);
l=1;
while (l<minPD+1)
    {
    ca=Cifra_en_posicion_PD(l);
    cb=b.Cifra_en_posicion_PD(l);
    if (ca<cb) return si;
      else if (ca>cb) return no;
    l++;
    };
if (nPD<b.nPD) return si;
if (nPD>b.nPD) return no;
// Son iguales
return NO;
};

int Numero::operator <=(Numero &b)
{
if (signo!=b.signo)
   {
   if (signo==negativo) return SI;
   return NO;
   };
// Tienen el mismo signo
bool si=SI;
bool no=NO;
if (signo==negativo) // si es negativo el mas grande es el de menos valor
   {
   si=NO;
   no=SI;
   };
if (nPE<b.nPE) return si;
if (nPE>b.nPE) return no;
ulong l=nPE;
Cifra ca,cb;
while (l>0)
    {
    ca=Cifra_en_posicion_PE(l-1);
    cb=b.Cifra_en_posicion_PE(l-1);
    if (ca<cb) return si;
      else if (ca>cb) return no;
    l--;
    };

ulong minPD=MINl(nPD,b.nPD);
l=1;
while (l<minPD+1)
    {
    ca=Cifra_en_posicion_PD(l);
    cb=b.Cifra_en_posicion_PD(l);
    if (ca<cb) return si;
      else if (ca>cb) return no;
    l++;
    };
if (nPD<b.nPD) return si;
if (nPD>b.nPD) return no;
// Son iguales
return SI;
};

int Numero::operator >(Numero &b)
{
if (signo!=b.signo)
   {
   if (signo==negativo) return NO;
   return SI;
   };
// Tienen el mismo signo
bool si=SI;
bool no=NO;
if (signo==negativo) // si es negativo el mas grande es el de menos valor
   {
   si=NO;
   no=SI;
   };
if (nPE<b.nPE) return no;
if (nPE>b.nPE) return si;
ulong l=nPE;
Cifra ca,cb;
while (l>0)
    {
    ca=Cifra_en_posicion_PE(l-1);
    cb=b.Cifra_en_posicion_PE(l-1);
    if (ca<cb) return no;
      else if (ca>cb) return si;
    l--;
    };

ulong minPD=MINl(nPD,b.nPD);
l=1;
while (l<minPD+1)
    {
    ca=Cifra_en_posicion_PD(l);
    cb=b.Cifra_en_posicion_PD(l);
    if (ca<cb) return no;
      else if (ca>cb) return si;
    l++;
    };
if (nPD>b.nPD) return si;
if (nPD<b.nPD) return no;
// Son iguales
return NO;
};

int Numero::operator >=(Numero &b)
{
if (signo!=b.signo)
   {
   if (signo==negativo) return NO;
   return SI;
   };
// Tienen el mismo signo
bool si=SI;
bool no=NO;
if (signo==negativo) // si es negativo el mas grande es el de menos valor
   {
   si=NO;
   no=SI;
   };
if (nPE<b.nPE) return no;
if (nPE>b.nPE) return si;
ulong l=nPE;
Cifra ca,cb;
while (l>0)
    {
    ca=Cifra_en_posicion_PE(l-1);
    cb=b.Cifra_en_posicion_PE(l-1);
    if (ca<cb) return no;
      else if (ca>cb) return si;
    l--;
    };

ulong minPD=MINl(nPD,b.nPD);
l=1;
while (l<minPD+1)
    {
    ca=Cifra_en_posicion_PD(l);
    cb=b.Cifra_en_posicion_PD(l);
    if (ca<cb) return no;
      else if (ca>cb) return si;
    l++;
    };
if (nPD>b.nPD) return si;
if (nPD<b.nPD) return no;
// Son iguales
return SI;
};

int Numero::operator ==(Numero &b)
{
if (signo!=b.signo) return NO;
if (nPE!=b.nPE) return NO;
if (nPD!=b.nPD) return NO;
ulong l=nPE;
Cifra ca,cb;
while (l>0)
    {
    ca=Cifra_en_posicion_PE(l-1);
    cb=b.Cifra_en_posicion_PE(l-1);
    if (ca!=cb) return NO;
    l--;
    };
l=1;
while (l<nPD+1)
    {
    ca=Cifra_en_posicion_PD(l);
    cb=b.Cifra_en_posicion_PD(l);
    if (ca!=cb) return NO;
    l++;
    };
return SI;
};

int Numero::operator !=(Numero &b)
{
if ((*this)==b) return NO;
return SI;
};

void Numero::Nuevo_tamano(ulong nuevoPE,ulong nuevoPD)
{
Nuevo_tamano_PE(nuevoPE);
Nuevo_tamano_PD(nuevoPD);
};

void Numero::Nuevo_tamano_PE(ulong new_num)
{
if (new_num==maxPE) return;
if (new_num==0) return;
//if (new_num==0) new_num=1;  // Por reservar algo siempre
Cifra *pe=(Cifra *)fmalloc(new_num,IDfmPE);
Cifra *psrc=PE;
Cifra *ptgt=pe;
ulong l;
if (new_num>maxPE)
   {
   l=new_num-maxPE;
   while (l>0)
      {
      *ptgt=0;
      ptgt++;
      l--;
      };
   l=maxPE;
   }
  else
   {
   l=new_num;
   psrc+=maxPE-new_num;
   };
while (l>0)
   {
   *ptgt=*psrc;
   psrc++;
   ptgt++;
   l--;
   };
if (PE!=NULL) ffree(PE);
PE=pe;
maxPE=new_num;
};

void Numero::Nuevo_tamano_PD(ulong new_num)
{
if (new_num==maxPD) return;
//if (new_num==0) new_num=1;  // Por reservar algo siempre
if (new_num==0) return;
Cifra *pd=(Cifra *)fmalloc(new_num,IDfmPD);
Cifra *psrc=PD;
Cifra *ptgt=pd;
ulong l;
if (new_num>maxPD)
   {
   l=maxPD;
   while (l>0)
      {
      *ptgt=*psrc;
      psrc++;
      ptgt++;
      l--;
      };
   l=new_num-maxPD;
   while (l>0)
      {
      *ptgt=0;
      ptgt++;
      l--;
      };
   }
  else
   {
   l=new_num;
   while (l>0)
      {
      *ptgt=*psrc;
      psrc++;
      ptgt++;
      l--;
      };
   };

if (PD!=NULL) ffree(PD);
PD=pd;
maxPD=new_num;

};

void Numero::operator <<=(ulong n)
{
Cifra *TGT=PE; // Estas tres variables estan para no tener que llamar a
ulong maxTGT=maxPE; // Nuevo_Tamano_PE, asi a la vez reserva la memoria
bool nuevoTGT=NO;   // necesaria y lo copia en su sitio en vez de reservar memoria
                    // copiarla y despues moverla a su sitio
if (maxPE<nPE+n)
   {
   nuevoTGT=SI;
   maxTGT=nPE+n;
   TGT=(Cifra *)fmalloc(maxTGT,IDfmPE);
   };
Cifra *ptgt;
Cifra *psrc;
ptgt=TGT+maxTGT-(nPE+n);
psrc=PE+maxPE-nPE;
ulong l=nPE;
while (l>0)
    {
    *ptgt=*psrc;
    ptgt++;
    psrc++;
    l--;
    };
// Ahora coge las cifras de la parte decimal para copiarlas en la parte entera
l=MIN(n,nPD);
psrc=PD;
while (l>0)
    {
    *ptgt=*psrc;
    ptgt++;
    psrc++;
    l--;
    };
if (n>=nPD)
   {          // Los decimales se han copiado (si habia) y ahora hay que rellenar con 0s
   l=n-nPD;
   while (l>0)
       {
       *ptgt=0;
       ptgt++;
       l--;
       };
   nPD=0;
   }
  else    // Los decimales que han quedado tambien hay que desplazarlos ...
   {
   l=nPD-n;
   ptgt=PD;
   while (l>0)
       {
       *ptgt=*psrc;
       ptgt++;
       psrc++;
       l--;
       };
   nPD=nPD-n;
   };
nPE+=n;
if (nuevoTGT==SI)
    {
    if (PE!=NULL) ffree(PE);
    PE=TGT;
    maxPE=maxTGT;
    };
Minimiza_PE();
};

void Numero::operator >>=(ulong n)
{
Cifra *TGT=PD; // Estas tres variables estan para no tener que llamar a
ulong maxTGT=maxPD; // Nuevo_Tamano_PE, asi a la vez reserva la memoria
bool nuevoTGT=NO;   // necesaria y lo copia en su sitio en vez de reservar memoria
                    // copiarla y despues moverla a su sitio
if (maxPD<nPD+n)
   {
   nuevoTGT=SI;
   maxTGT=nPD+n;
   TGT=(Cifra *)fmalloc(maxTGT,IDfmPD);
   };
Cifra *ptgt;
Cifra *psrc;
ptgt=TGT+(nPD+n-1);
psrc=PD+(nPD-1);
ulong l=nPD;
while (l>0)
    {
    *ptgt=*psrc;
    ptgt--;
    psrc--;
    l--;
    };
// Ahora coge las cifras de la parte entera para copiarlas en la parte decimal
l=MIN(n,nPE);
psrc=PE+maxPE-1;
while (l>0)
    {
    *ptgt=*psrc;
    ptgt--;
    psrc--;
    l--;
    };
if (n>=nPE)
   {          // Los enteros se han copiado (si habia) y ahora hay que rellenar con 0s
   l=n-nPE;
   while (l>0)
       {
       *ptgt=0;
       ptgt--;
       l--;
       };
   nPE=0;
   }
  else    // Los enteros que han quedado tambien hay que desplazarlos ...
   {
   l=nPE-n;
   ptgt=PE+maxPE-1;
   while (l>0)
       {
       *ptgt=*psrc;
       ptgt--;
       psrc--;
       l--;
       };
   nPE=nPE-n;
   };
nPD+=n;
if (nuevoTGT==SI)
    {
    if (PD!=NULL) ffree(PD);
    PD=TGT;
    maxPD=maxTGT;
    };
Minimiza_PD();
};

Cifra Numero::CifraxCifra(Cifra a,Cifra b,Cifra &Carry)
{
Double_Cifra tmp;
tmp=(a*b)+Carry;
Double_Cifra Carrytmp=0;
while ((Carrytmp+1)*Base<=tmp)
   {
   Carrytmp++;
   };
Carry=Carrytmp;
tmp-=Carrytmp*Base;
return tmp;
};

void Numero::Multiplica_por_cifra(Cifra c,Numero *Resultado)
{
if (c==0)
   {
   Resultado->Equal(0);
   return ;
   }
if (c==1)
   {
   (*Resultado)=(*this);
   return;
   };

Resultado->Nuevo_tamano_PE(nPE+1);
ulong l=0;
Cifra *pe=PE+maxPE-1;
Cifra *ptgt=Resultado->PE+Resultado->maxPE-1;
Cifra carry=0;
while (l<nPE)
   {
   *ptgt=CifraxCifra(c,*pe,carry);
   ptgt--;
   pe--;
   l++;
   };
if (carry==0)
   {
   Resultado->nPE=nPE;
   }
  else
   {
   *ptgt=carry;
   Resultado->nPE=nPE+1;
   };
};

Numero Numero::Multiplica_normal(Numero &b)
{
Cifra *pB=b.PE+b.maxPE-1;
Inicializa_tabla_multiplicar();
ulong l=0;
Numero Temporal;//(maxPE+b.maxPE,0);
Numero Resultado;
short i;
while (l<b.nPE)
   {
   i=*pB;
   if (Tabla_multiplicar[i]==NULL)
       {
       Tabla_multiplicar[i]=new Numero(nPE+1,0);
       Multiplica_por_cifra(*pB,Tabla_multiplicar[i]);
       };
   Temporal=*Tabla_multiplicar[i];
   Temporal<<=l;
   Resultado.Incrementa(Temporal);
   pB--;
   l++;
   };
return Resultado;
};

Numero Numero::operator *(Numero &b_)
{
// Todos estos desplazamientos estn ah para tratar bien el producto con
//   decimales
Numero b(b_);
ulong oldthisnPD=nPD;
ulong oldbnPD=b.nPD;
operator <<=(nPD);
b<<=b.nPD;
Numero Resultado(Multiplica_normal(b));
Resultado.Err=No_hay_error;
// Deja los dos objetos como estaban
b>>=oldbnPD;
operator >>=(oldthisnPD);
Resultado>>=oldbnPD+oldthisnPD;
if (b.signo!=signo)
   Resultado.signo=negativo;
return Resultado;
};

Numero Numero::Factorial()
{
Numero Resultado(1);
Numero Tmp(1);
Numero Uno(1);
while (Tmp<(*this))
   {
   Resultado=Resultado.Multiplica_normal(Tmp);
                         // Se pueden hacer suposiciones respecto a los signos
                         //  que es la unica diferencia entre usar el operador *
                         //  y la funcion Multiplica_normal
   Tmp++;
   };
Resultado=(Resultado*Tmp);
return Resultado;
};

Numero Numero::operator ++(void)
{
if (signo==negativo)
   {
   if ((nPE==1)&&(*(PE+maxPE-1)==1)) // (-1.xxx)++  que es 0.xxx, lo hacemos "cableado"
      {
      nPE=0;
      if (nPD==0) signo=positivo;  // es (-1)++  =  0
         else signo=negativo;
      return (*this);
      };
   signo=positivo;
   if (nPE==0)
      {
      operator--();
      signo=positivo;
      }
     else
      {
      operator--();
      signo=negativo;
      };
   return (*this);
   };
if (nPE==0)  // El caso de 0.xxx se trata aparte porque hay que reservar memoria y aumentar nPE
   {
   Nuevo_tamano_PE(1);
   Cifra *pe=PE;
   pe+=maxPE-1;
   (*pe)++;
   nPE=1;
   return (*this);
   };
Cifra *pe=PE;
pe+=maxPE-1;
(*pe)++;
ulong l=maxPE;
while ((*pe>=Base)&&(l>1))
   {
   *pe-=Base;         // Y me llevo un carry de 1
   pe--;
   l--;
   (*pe)++;
   };
if ((l==1)&&(*pe>=Base))  // Hay que ampliar en una cifra
   {
   Nuevo_tamano_PE(maxPE+1);
   *PE=1;             // El carry es de 1 seguro
   (*(PE+1))-=Base;
   nPE++;
   };
return (*this);
};

Numero Numero::operator --(void)
{
if (nPE==0) // Si es 0.xxx al hacerle el -- los decimales cambian, por lo que
            // hay que llamar al - usual
   {
   Numero Uno(1);
   (*this)=operator -(Uno); // Se podria poner operator -(1); pero el compilador da
             //  un warning porque pasa un temporal por referencia, y no me gustan
             //  los warnings   
   return (*this);
   };
if (signo==negativo)
   {
   signo=positivo;
   operator++();
   signo=negativo;
   return (*this);
   };
Cifra *pe=PE;
pe+=maxPE-1;
bool fin=NO;
// En el siguiente bucle no hace falta tener en cuenta a nPE porque se puede
//  suponer que las variables son consistentes y entonces se comprueba que
//  SIEMPRE acaba, excepto en el caso del 0 pero este caso ya se comprob al
//  principio de la funcin
while (!fin)
   {
   if (*pe!=0)
      {
      (*pe)--;
      fin=SI;
      }
     else
      {
      (*pe)=Base-1;
      };
   pe--;
   };
return (*this);
};

Numero Numero::operator ++(int)
{
Numero Antes_de_cambios(*this);
operator++();
return Antes_de_cambios;
};

Numero Numero::operator --(int)
{
Numero Antes_de_cambios(*this);
operator--();
return Antes_de_cambios;
};

void Numero::Anade_a_la_derecha_en_PE(Cifra c)
{
if ((nPE==0)&&(c==0)) return; // Esto ocurre si ya es 0 y se le va a aadir un 0

Nuevo_tamano_PE(nPE+1);
ulong l=nPE;
Cifra *psrc=PE+(maxPE-nPE);
Cifra *ptgt=PE+(maxPE-nPE-1);
while (l>0)
   {
   *ptgt=*psrc;
   ptgt++;
   psrc++;
   l--;
   };
*ptgt=c;
nPE++;
};

void Numero::Anade_a_la_derecha_en_PD(Cifra c)
{
Nuevo_tamano_PD(nPD+1);
Cifra *ptgt=PD+nPD;
*ptgt=c;
nPD++;
};

void Numero::Inicializa_tabla_multiplicar(void)
{
ulong l=0;
if (Tabla_multiplicar!=NULL)
    {
    if ((*Tabla_multiplicar[1])==(*this)) return;
    while (l<Base)
       {
       if (Tabla_multiplicar[l]!=NULL)
           {
           delete Tabla_multiplicar[l];
           Tabla_multiplicar[l]=NULL;
           };
       l++;
       };
//    delete Tabla_multiplicar;
    }
   else
    {
    Tabla_multiplicar=new Numero*[Base];
    while (l<Base)
       {
       Tabla_multiplicar[l]=NULL;
       l++;
       };
    };
//   Tabla_multiplicar[0]=new Numero;
Tabla_multiplicar[1]=new Numero(*this);// Aqui se copia a si mismo para poder
       // reusar la tabla si se hacen varias multiplicaciones seguidas con el mismo
       // numero. Asi sabe despues si se puede reusar o no (con la comparacion del
       // principio de esta funcion
};

Cifra Numero::Divide_N_entre_N(Numero *r,Numero &divisor)
{
if (*r<divisor) return 0;
if ((*r)>=(*divisor.Tabla_multiplicar[Base-1]))
   {
   (*r)=(*r)-(*divisor.Tabla_multiplicar[Base-1]);
   return Base-1;
   };

// Esta funcion hace una "Busqueda binaria" del resultado, y va guardando
// algunos resultados en un array por si en una llamada posterior a esta
// funcion lo necesita, poder ir mas rapido.
Cifra a=1;     // Cota inferior
Cifra b=Base-1;  // Cota superior
short i;
while (b-a>1)
   {
   i=(a+b)/2;
   if (divisor.Tabla_multiplicar[i]==NULL)
      {
      divisor.Tabla_multiplicar[i]=new Numero(divisor.nPE+1,0);
      divisor.Multiplica_por_cifra(i,divisor.Tabla_multiplicar[i]);
      };
   if ((*divisor.Tabla_multiplicar[i])<(*r))
       {
       a=i;
       }
      else
       if ((*divisor.Tabla_multiplicar[i])>(*r))
           {
           b=i;
           }
          else // Es igual
           {
           r->nPE=0;
           return i;
           };
   };
(*r)=(*r)-(*divisor.Tabla_multiplicar[a]);
return a;
};

Numero Numero::operator /(Numero &divisor)
{
Numero Cociente;
if ((divisor.nPE==0)&&(divisor.nPD==0)) {Cociente.Err=Division_x_cero;return Cociente;}
sign signodivisor=divisor.signo;
divisor.signo=positivo;
divisor.Minimiza_PD();
ulong oldbnPD=divisor.nPD;
ulong maxnPD=MAX(divisor.nPD,nPD);
ulong maxnPE=MAX(divisor.nPE,nPE);
divisor<<=divisor.nPD;

divisor.Inicializa_tabla_multiplicar();
// Las tres siguientes lineas terminan de inicializar la tabla de multiplicar
//  ya que al dividir es necesario que est preparado la multiplicacion por
//  la ultima cifra posible de la Base
short i=Base-1;
if (divisor.Tabla_multiplicar[i]==NULL) divisor.Tabla_multiplicar[i]=new Numero(nPE+1,0);
divisor.Multiplica_por_cifra(i,divisor.Tabla_multiplicar[i]);

Numero Temporal;
ulong l=nPE;
Cifra *p=PE+(maxPE-nPE);
Cifra c;
// Empieza cogiendo de la parte entera ...
while (l>0)
   {
   Temporal.Anade_a_la_derecha_en_PE(*p);
   c=Divide_N_entre_N(&Temporal,divisor);
   Cociente.Anade_a_la_derecha_en_PE(c);
   p++;
   l--;
   };
//  ... y como se nos ha terminado la parte entera, debe continuar cogiendo
//   de la parte decimal (mientras queden numeros) ...
p=PD;
l=nPD;
while (l>0)
   {
   Temporal.Anade_a_la_derecha_en_PE(*p);
   c=Divide_N_entre_N(&Temporal,divisor);
   Cociente.Anade_a_la_derecha_en_PD(c);
   p++;
   l--;
   };

// ... y como tambien se nos ha terminado la parte decimal, ahora va cogiendo
//   ceros . Cuidado con los numeros periodicos !!!

//   La comparacion de Cociente.nPD se hace con maxnPD+maxnPE+MAX_CIFRAS_DEC en
//   lugar de con MAX_CIFRAS_DEC porque as tiene tanta o mas precision que los
//   operandos ya que en otro caso el error cometido sera muy grande en ciertos
//   casos.
//   Como 1/0.00000000000000000000000000000000000000003 , que da como resultado
//          con MAX_CIFRAS_DEC    :33333333333333333333333333333300000000000
//con maxnPD+maxnPE+MAX_CIFRAS_DEC:33333333333333333333333333333333333333333.333333333333333333333333333333
//  o sea , que asi se evitan errores que pueden llegar a 33333333334
//                                                       (miles de millones)

while ((Temporal.nPE!=0)&&(Cociente.nPD<maxnPD+maxnPE+MAX_CIFRAS_DEC))
   {
   Temporal.Anade_a_la_derecha_en_PE(0);
   c=Divide_N_entre_N(&Temporal,divisor);
   Cociente.Anade_a_la_derecha_en_PD(c);
   l--;
   };

if ((Cociente.nPD==maxnPD+maxnPE+MAX_CIFRAS_DEC)&&(Temporal.nPE!=0)) Cociente.Err=Cociente_periodico;
divisor>>=oldbnPD;
Cociente<<=oldbnPD;

if (signo!=signodivisor) Cociente.signo=negativo;
divisor.signo=signodivisor;

return Cociente;
};

void Numero::Cambia_signo(void)
{
if ((nPE==0)&&(nPD==0)) {signo=positivo;return;};
if (signo==positivo) signo=negativo;
    else signo=positivo;
};

Numero Numero::Valor_Absoluto(void)
{
Numero Resultado(*this);
Resultado.signo=positivo;
return Resultado;
};

ulong Numero::Longitud(void)
{
  return nPE+nPD+1+((signo==positivo)?(0):(1));
};

void Numero::Precision(ulong cuantos_dejar)
{
  if (nPD>cuantos_dejar) 
      {
      nPD=cuantos_dejar;
      if (Err==Cociente_periodico) Err=No_hay_error;
      };
 
};
