/**********************************************************************
    numero.h
  
    RealNumbers  1.0
    Copyright (C) 1997  Antonio Larrosa Jimenez

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
***********************************************************************/

#ifndef _NUMERO_H
#define _NUMERO_H

#include "dattypes.h"
#include "errores.h"
#include "numconst.h"
#include <stream.h>

class Numero
{
friend class NumeroZ;
friend class NumeroQ;
friend class NumeroE;

private:
Tipo_Error Err;

sign signo; // El signo del numero

Cifra Base;

Cifra *PE; // Parte entera
Cifra *PD; // Parte Decimal

ulong nPE; // numero de cifras de la parte entera
ulong nPD; // numero de cifras de la parte decimal

ulong maxPE; // El numero de cifras RESERVADAS en la parte entera
ulong maxPD; // El numero de cifras RESERVADAS en la parte decimal

Cifra Caracter_a_Cifra(char c,bool &ok); // Convierte de ASCII a byte y devuelve ok=NO si no es una cifra permitida (por ejemplo un signo $)
char Cifra_a_Caracter(Cifra c); // Convierte de ASCII a byte

void Nuevo_tamano(ulong nuevoPE,ulong nuevoPD);
void Nuevo_tamano_PE(ulong new_units);
void Nuevo_tamano_PD(ulong new_units);

void Incrementa(Numero &b);           // Es parecido al significado de += pero
                                      //  solo incrementa la parte entera
Numero Multiplica_normal(Numero &b);  // Multiplica (*this) por b y devuelve
                                      // el resultado, pero ATENCION :
                                      // solo tiene en cuenta la parte entera
Cifra CifraxCifra(Cifra a,Cifra b,Cifra &Carry);
void Multiplica_por_cifra(Cifra c,Numero *Resultado);

Numero **Tabla_multiplicar;
// Al inicializarlo se puede usar como un array que tiene Base elementos
//  es decir, como si la definicion fuera Numero *Tabla_multiplicar[Base];
//  Al hacer una division a mano, multiplicamos el divisor por varios numeros
//  para obtener un numero que aadir al cociente, si el numero es grande,
//  se repiten las multiplicaciones, por qu no guardarlas y as no
//  hay que repetir la misma operacion?. Para eso sirve esta variable
void Inicializa_tabla_multiplicar(void);
Cifra Divide_N_entre_N(Numero *r,Numero &divisor);

void Anade_a_la_derecha_en_PE(Cifra c);
     //Basicamente, hacer a.Anade_a_la_derecha(c); es como hacer :
                                  //   a<<=1;
                                  //   *(PE+maxPE-1)=c;
void Anade_a_la_derecha_en_PD(Cifra c);
     // Lo mismo pero en la parte decimal

public:
Numero();
Numero(long l);
Numero(Numero &a);
Numero(NumeroZ &a);
Numero(ulong pe,ulong pd,bool inicializa=SI); // Tamaos de PE y PD y si debe
                                              //  rellenarlos con 0
~Numero();

ulong NPE(void) {return nPE;};
ulong NPD(void) {return nPD;};
Cifra Cifra_en_posicion_PE(ulong l); // Desde 0 (LSB) al infinito (MSB)
Cifra Cifra_en_posicion_PD(ulong l); // Desde 1 (MSB) al infinito (LSB)

sign Signo(void) {return signo;};
Numero operator =(long l);            
Numero Equal(Cifra c);
Numero operator =(char *s);    // Adquiere el numero de una cadena
Numero operator =(Numero a);
void Convierte_a_cadena(char *s);   // Copia el numero a una cadena

friend ostream &operator<<(ostream& s,Numero arg);

void Cambia_signo(void);
Numero Valor_Absoluto(void);

Numero operator ++(void);
Numero operator --(void);
Numero operator ++(int);
Numero operator --(int);

void operator <<=(ulong l); /* Esta funcion se puede usar para multiplicar
                             por la base, por ejemplo si la base es 10 y
                             se quiere multiplicar por 100, pues se pone
                                      numero<<=2;
                             que es la misma sentencia que se pondria si la
                             base fuera 2 y se quisiera multiplicar por 4 */
void operator >>=(ulong l); /* Lo mismo que arriba, simplemente desplaza a la
                             derecha o divide */

Numero operator +(Numero &b);
Numero operator -(Numero &b);
Numero operator -(void);
Numero operator *(Numero &b);
Numero operator /(Numero &b);

Numero Factorial(void);

int operator <(Numero &b);
int operator >(Numero &b);
int operator <=(Numero &b);
int operator >=(Numero &b);
int operator ==(Numero &b);
int operator !=(Numero &b);

void Minimiza_PE();
void Minimiza_PD();

ulong Longitud(void);  // Da la cantidad de caracteres que ocupa la cadena
                  // (se usa para alocatear la memoria antes de llamar a Convierte_a_cadena)

void Precision(ulong cuantos_decimales_dejar);

void SetErrorFlag(Tipo_Error e) {Err=e;};
};

#endif
