/**********************************************************************
    numeroe.h
  
    RealNumbers  1.0
    Copyright (C) 1997  Antonio Larrosa Jimenez

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
***********************************************************************/
#ifndef _NUMEROE_H
#define _NUMEROE_H

#include "dattypes.h"
#include "errores.h"
#include "numero.h"
#include <stream.h>
#include "numeroq.h"

class NumeroE
{
friend class Numero;
friend class NumeroQ;


private:
Tipo_Error Err;

Numero *NumRed;  // El numero redondeado
Numero *Error;   // Una cota superior del valor absoluto del error


public:

NumeroE();
NumeroE(long l);
NumeroE(NumeroZ &a);
NumeroE(NumeroQ &a);
NumeroE(NumeroE &a);
NumeroE(Numero  &a);

~NumeroE();

void Cambia_signo(void);

NumeroE Equal(Cifra c);
NumeroE operator =(long l);
NumeroE operator =(char *s);    // Adquiere el numero de una cadena
NumeroE operator =(NumeroE a);
void Convierte_a_cadena(char *s);   // Copia el numero a una cadena

friend ostream &operator<<(ostream& s,NumeroE arg);
NumeroE operator ++(void);
NumeroE operator --(void);
NumeroE operator ++(int);
NumeroE operator --(int);

void operator <<=(ulong l); /* Esta funcion se puede usar para multiplicar
                             por la base, por ejemplo si la base es 10 y
                             se quiere multiplicar por 100, pues se pone
                                      numero<<=2;
                             que es la misma sentencia que se pondria si la
                             base fuera 2 y se quisiera multiplicar por 4 */
void operator >>=(ulong l); /* Lo mismo que arriba, simplemente desplaza a la
                             derecha o divide */

NumeroE operator +(NumeroE &a);
NumeroE operator -(NumeroE &b);
NumeroE operator -(void);
NumeroE operator *(NumeroE &b);
NumeroE operator /(NumeroE &b);
NumeroE Divide_dos_numeros(Numero &a,Numero &b);

/*
En las siguientes funciones ocurre que no se puede devolver un SI rotundo
en ciertos casos como por ejemplo,si A=2.5+-0.5 [2.0,3.0] y B=2.9+-0.3 [2.6,3.2]
en este caso no se puede decir que A sea igual que B ya que no se sabe el valor
exacto, por eso hay que tener en cuenta que ademas de devolver un SI y un NO,
en muchos casos se devolver un QUIZAS que es lo mismo que un SI posible, por
tanto el SI y el NO se devolver slo cuando sean un SI o un NO seguro
Asi, en el ejemplo anterior si se compara la igualdad, devolver un QUIZAS.
Ver mas explicaciones en numeroe.cc antes de la implementacion de estos 
operadores
  */

int operator <(NumeroE &b);
int operator >(NumeroE &b);
int operator <=(NumeroE &b);
int operator >=(NumeroE &b);
int operator ==(NumeroE &b);
int operator !=(NumeroE &b);

bool Es_positivo(void);
bool Es_negativo(void);

ulong Longitud(void);

Numero Minimo_valor_posible(void);
Numero Maximo_valor_posible(void);
bool Puede_ser(Numero &a);
                 // Devuelve SI si 'a' esta dentro del rango de posibles
                 // valores del objeto en otro caso devuelve NO

void Precision(ulong cuantos_decimales_dejar);
void SetErrorFlag(Tipo_Error e) {Err=e;};
};


#endif
