/**********************************************************************
   numeroq.cc   --  La clase "NumeroQ" que implementa un numero grande de Q
                    el conjunto de los numeros racionales

 
    RealNumbers  1.0
    Copyright (C) 1997  Antonio Larrosa Jimenez

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
***********************************************************************/
 

#include "numeroq.h"
#include <malloc.h>
#include "newdebug.h"
#include "numfunc.h"

NumeroQ::NumeroQ()
{
Err=No_hay_error;
signo=positivo;
Numerador=new NumeroZ;
Denominador=new NumeroZ(1);
};

NumeroQ::NumeroQ(long l)
{
Err=No_hay_error;
signo=(l<0)?(negativo):(positivo);
Numerador=new NumeroZ(l);
Denominador=new NumeroZ(1);
};

NumeroQ::NumeroQ(NumeroQ &a)
{
Err=a.Err;
signo=a.signo;
Numerador=new NumeroZ((*a.Numerador));
Denominador=new NumeroZ((*a.Denominador));
};

NumeroQ::NumeroQ(NumeroZ &num,NumeroZ &den,sign elsigno)
{
Err=No_hay_error;
signo=elsigno;
Numerador=new NumeroZ(num);
Denominador=new NumeroZ(den);
};

NumeroQ::NumeroQ(Numero &a)
{
Err=a.Err;
signo=a.signo;
Numero copia(a);
copia<<=copia.nPD;
Numerador=new NumeroZ(copia);
NumeroZ den(1);
den<<=a.nPD;
Denominador=new NumeroZ(den);
Reduce();
};

NumeroQ::~NumeroQ()
{
if (Numerador!=NULL)
   {
   delete Numerador;
   Numerador=NULL;
   };
if (Denominador!=NULL)
   {
   delete Denominador;
   Denominador=NULL;
   };
};

void NumeroQ::Reduce(void)
{
if (Numerador->signo!=Denominador->signo)
   {
   Cambia_signo();
   Numerador->signo=positivo;
   Denominador->signo=positivo;
   }
  else
   {
   Numerador->signo=positivo;
   Denominador->signo=positivo;
   };
if (Numerador->nPE==0) // Si el numerador es 0  ...
   {
   Denominador->Equal(1);
   return;
   };
if (Denominador->nPE==0) // Si el denominador es 0  ...
   {
   Err=Division_x_cero;
   return;
   };
// Ya estn los signos bien puestos y ni el numerador ni el denominador es 0
NumeroZ Cociente1;
NumeroZ Resto1;
Numerador->Division_da_cociente_y_resto((*Denominador),Cociente1,Resto1); 
	       // Llamo a esta funcion en vez de a % para aprovechar operaciones
if (Resto1.nPE==0)
   {
   (*Numerador)=Cociente1;
   Denominador->Equal(1);
   return;
   };

/* Esta era la antigua forma de sacar los factores comunes, pero despues viene
	la nueva forma, que es infinitamente mas rpida ...


NumeroZ Cociente2;
NumeroZ Resto2;
NumeroZ Divisor;
++Divisor;
++Divisor;
// Divisor = 2, pero lo hago asi para que no haya fallos si la base es 2
bool fin=NO;
// Primero divido entre 2 numerador y denominador, todas las veces que se pueda
while (!fin)
   {
   Numerador->Division_da_cociente_y_resto(Divisor,Cociente1,Resto1);
   Denominador->Division_da_cociente_y_resto(Divisor,Cociente2,Resto2);
   if ((Resto1.nPE==0)&&(Resto2.nPE==0)) // Si los dos son multiplos de 2
      {
      (*Numerador)=Cociente1;
      (*Denominador)=Cociente2;
      }
     else  // Alguno de los dos ya no es multiplo de 2
      fin=SI;
   };
++Divisor;		// Divisor = 3
NumeroZ MinNumDen;
if (Numerador<Denominador)
   {
   MinNumDen=Cociente1; // MinNumDen=Numerador/2;
   ++MinNumDen;
   }
  else // Denominador<Numerador
   {
   MinNumDen=Cociente2; // MinNumDen=Denominador/2;
   ++MinNumDen;
   };
while (Divisor<MinNumDen)
   {
   Numerador->Division_da_cociente_y_resto(Divisor,Cociente1,Resto1);
   Denominador->Division_da_cociente_y_resto(Divisor,Cociente2,Resto2);
   if ((Resto1.nPE==0)&&(Resto2.nPE==0))
      {
      (*Numerador)=Cociente1;
      (*Denominador)=Cociente2;
      }
     else
      {
// Ahora va de 2 en 2 porque los numeros pares no pueden dividir a Numerador y
// Denominador gracias al primer bucle de la funcion
      ++Divisor;
      ++Divisor;
      };
   };
 ... Y ahora la velocidad*/

NumeroZ Mcd;
Mcd=(*Numerador).MCD(*Denominador);
*Numerador=(*Numerador)/Mcd;
*Denominador=(*Denominador)/Mcd;
};

void NumeroQ::Cambia_signo(void)
{
if (signo==positivo) signo=negativo;
 else signo=positivo;
};

NumeroQ NumeroQ::operator =(long l)
{
*Numerador=l;
*Denominador=1L;
return (*this);
};

NumeroQ NumeroQ::operator =(char *s)
{
Err=No_hay_error;
if (Es_un_numero_correcto(s,Numerador->Base,"-/")==NO) return (*this);
signo=positivo;
char oldchar;
char *pc=s;
while ((*pc!=0)&&(*pc!='/')) pc++;
oldchar=*pc;
*pc=0;
Numerador=new NumeroZ();
(*Numerador)=s;
*pc=oldchar;
if (oldchar=='/')
   {
   pc++;
   Denominador=new NumeroZ();
   (*Denominador)=pc;
   if (Denominador->nPE==0)
       {
       Err=Division_x_cero;
       };
   }
  else
   {
   Denominador=new NumeroZ(1);
   };
Reduce();
return (*this);
};

NumeroQ NumeroQ::operator =(NumeroQ a)
{
Err=a.Err;
if (Numerador!=NULL)
   {
   delete Numerador;
   Numerador=NULL;
   };
if (Denominador!=NULL)
   {
   delete Denominador;
   Denominador=NULL;
   };
signo=a.signo;
if (a.Numerador!=NULL) Numerador=new NumeroZ(*a.Numerador);
if (a.Denominador!=NULL) Denominador=new NumeroZ(*a.Denominador);

return (*this);
};

ostream &operator<<(ostream& s,NumeroQ arg)
{
 if (arg.Denom_es_cero()) arg.Err=Division_x_cero;
 switch (arg.Err)
     {
     case (Division_x_cero) : s << "Division por cero" ; break;
     default :
           char *pstr=(char *)fmalloc((*arg.Numerador).Longitud()+(*arg.Denominador).Longitud()+3L,0);
           arg.Convierte_a_cadena(pstr);
           s << pstr;
           ffree(pstr);
     };
  return s;
};

void NumeroQ::Convierte_a_cadena(char *s)
{
Reduce();
char *pc=s;
if (signo==negativo)
    {
    *pc='-';
    pc++;
    };
Numerador->Convierte_a_cadena(pc);
while (*pc!=0) pc++;

NumeroZ Uno(1);
if ((*Denominador)==Uno)
   {
   *pc=0;
   }
  else
   {
   *pc='/';
   pc++;
   Denominador->Convierte_a_cadena(pc);
   };
};

void NumeroQ::operator<<=(ulong l)
{
(*Numerador)<<=l;
};

void NumeroQ::operator>>=(ulong l)
{
(*Denominador)<<=l;
};

NumeroQ NumeroQ::operator ++(void)
{
if (signo==negativo)
   {
   signo=positivo;
   operator --();
   signo=negativo;
   return (*this);
   };
(*Numerador)=(*Numerador)+(*Denominador);
return (*this);
};

NumeroQ NumeroQ::operator --(void)
{
if (signo==negativo)
   {
   signo=positivo;
   operator ++();
   signo=negativo;
   return (*this);
   };
(*Numerador)=(*Numerador)-(*Denominador);
return (*this);
};

NumeroQ NumeroQ::operator ++(int)
{
NumeroQ Antes_de_cambios(*this);
operator ++();
return Antes_de_cambios;
};

NumeroQ NumeroQ::operator --(int)
{
NumeroQ Antes_de_cambios(*this);
operator --();
return Antes_de_cambios;
};

NumeroQ NumeroQ::operator +(NumeroQ &b)
{
NumeroQ resultado;
if (signo!=b.signo)
   {
   if (b.signo==negativo)
      {
      b.signo=positivo;
      resultado=(*this)-(b);
      b.signo=negativo;
      }
     else
      {
      signo=positivo;
      resultado=b-(*this);
      signo=negativo;
      };
   return resultado;
   };
resultado.Numerador=new NumeroZ;
*resultado.Numerador=((*Numerador)*(*b.Denominador))+((*Denominador)*(*b.Numerador));
resultado.Denominador=new NumeroZ;
*resultado.Denominador=(*Denominador)*(*b.Denominador);
resultado.signo=signo;
return resultado;
};

NumeroQ NumeroQ::operator -(NumeroQ &b)
{
NumeroQ resultado;
if (signo!=b.signo)
   {
   if (signo==positivo)  //  (+a)-(-b) = (+a)+(+b) = (a+b)
      {
      b.signo=positivo;
      resultado=(*this)+b;
      b.signo=negativo;
      }
     else                //  (-a)-(+b) = (-a)+(-b) = -(a+b)
      {
      signo=positivo;
      resultado=(*this)+b;
      signo=negativo;
      resultado.signo=negativo;
      };
   return resultado;
   };
if (signo==positivo)  //Los dos son positivos
   {
   if ((*this)<b)
      {
      resultado=b-(*this);
      resultado.Cambia_signo();
      return resultado;
      };
   }
  else  // Los dos son negativos  : (-a)-(-b) = b-a
   {
      b.signo=positivo;
      signo=positivo;
      resultado=b-(*this);
      b.signo=negativo;
      signo=negativo;
      return resultado;
   };

resultado.Numerador=new NumeroZ;
*resultado.Numerador=((*Numerador)*(*b.Denominador))-((*Denominador)*(*b.Numerador));
resultado.Denominador=new NumeroZ;
*resultado.Denominador=(*Denominador)*(*b.Denominador);

return resultado;
};

NumeroQ NumeroQ::operator -(void)
{
NumeroZ Cero;
if (*Numerador==Cero) return (*this);
NumeroQ Resultado(*this);
Resultado.Cambia_signo();
return Resultado;
};

NumeroQ NumeroQ::operator *(NumeroQ &b)
{
NumeroQ Resultado;
Resultado.Numerador=new NumeroZ;
*Resultado.Numerador=(*Numerador)*(*b.Numerador);
Resultado.Denominador=new NumeroZ;
*Resultado.Denominador=(*Denominador)*(*b.Denominador);
if (signo!=b.signo) Resultado.signo=negativo;
return Resultado;
};

NumeroQ NumeroQ::operator /(NumeroQ &b)
{
NumeroQ Resultado;
Resultado.Numerador=new NumeroZ;
*Resultado.Numerador=(*Numerador)*(*b.Denominador);
Resultado.Denominador=new NumeroZ;
*Resultado.Denominador=(*Denominador)*(*b.Numerador);
if (signo!=b.signo) Resultado.signo=negativo;
if (b.Numerador->nPE==0) Err=Division_x_cero;
return Resultado;
};

int NumeroQ::operator <(NumeroQ &b)
{
if (signo!=b.signo)
   {
   if (signo==negativo) return SI;
   return NO;
   };
bool si=SI;
bool no=NO;
if (signo==negativo)
   {
   si=NO;
   no=SI;
   };
if ((*Denominador)==(*b.Denominador))
   {
   if ((*Numerador)<(*b.Numerador)) return si;
   return no;
   };
NumeroZ temp;
temp=(*b.Numerador)*(*Denominador);
if ((*Numerador)*(*b.Denominador)<temp) return si;
return no;
};

int NumeroQ::operator <=(NumeroQ &b)
{
if (signo!=b.signo)
   {
   if (signo==negativo) return SI;
   return NO;
   };
if ((*Denominador)==(*b.Denominador))
   {
   if (signo==positivo)
      {
      if ((*Numerador)<=(*b.Numerador)) return SI;
      }
     else
      {
      if ((*Numerador)>=(*b.Numerador)) return SI;
      };
   return NO;
   };
NumeroZ temp;
temp=(*b.Numerador)*(*Denominador);
if (signo==positivo)
   {
   if ((*Numerador)*(*b.Denominador)<=temp) return SI;
   }
  else
   {
   if ((*Numerador)*(*b.Denominador)>=temp) return SI;
   };
return NO;
};

int NumeroQ::operator >(NumeroQ &b)
{
if (signo!=b.signo)
   {
   if (signo==positivo) return SI;
   return NO;
   };
bool si=SI;
bool no=NO;
if (signo==negativo)
   {
   si=NO;
   no=SI;
   };
if ((*Denominador)==(*b.Denominador))
   {
   if ((*Numerador)>(*b.Numerador)) return si;
   return no;
   };
NumeroZ temp;
temp=(*b.Numerador)*(*Denominador);
if (((*Numerador)*(*b.Denominador))>temp) return si;
return no;
};

int NumeroQ::operator >=(NumeroQ &b)
{
if (signo!=b.signo)
   {
   if (signo==negativo) return SI;
   return NO;
   };
if ((*Denominador)==(*b.Denominador))
   {
   if (signo==positivo)
      {
      if ((*Numerador)>=(*b.Numerador)) return SI;
      }
     else
      {
      if ((*Numerador)<=(*b.Numerador)) return SI;
      };
   return NO;
   };
NumeroZ temp;
temp=(*b.Numerador)*(*Denominador);
if (signo==positivo)
   {
   if ((*Numerador)*(*b.Denominador)>=temp) return SI;
   }
  else
   {
   if ((*Numerador)*(*b.Denominador)<=temp) return SI;
   };
return NO;
};

int NumeroQ::operator ==(NumeroQ &b)
{
if (signo!=b.signo) return NO;
if ((*Denominador)==(*b.Denominador))
   {
   if ((*Numerador)==(*b.Numerador)) return SI;
   return NO;
   };
NumeroZ temp;
temp=(*b.Numerador)*(*Denominador);
return ((*Numerador)*(*b.Denominador)==(temp));
};

int NumeroQ::operator !=(NumeroQ &b)
{
if ((*this)==b) return NO;
return SI;
};

NumeroQ NumeroQ::Valor_Absoluto(void)
{
NumeroQ resultado(*this);
resultado.signo=positivo;
return resultado;
};

void NumeroQ::Precision(ulong)
{
};

