/**********************************************************************
    numeroq.h
  
    RealNumbers  1.0
    Copyright (C) 1997  Antonio Larrosa Jimenez

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
***********************************************************************/
#ifndef _NUMEROQ_H
#define _NUMEROQ_H

#include "dattypes.h"
#include "numconst.h"
#include "numeroz.h"
#include "numero.h"
#include "errores.h"

class NumeroQ
{
friend NumeroZ;
friend Numero;
friend class NumeroE;

private:
Tipo_Error Err;
sign signo; // El signo del numero

NumeroZ *Numerador;
NumeroZ *Denominador;

void Cambia_signo(void);

void Reduce(void);
public:
NumeroQ();
NumeroQ(long l);
NumeroQ(NumeroQ &a);
NumeroQ(NumeroZ &num,NumeroZ &den,sign elsigno=positivo);
NumeroQ(Numero &a);
~NumeroQ();

sign Signo(void) {return signo;};
NumeroQ operator =(long l);
NumeroQ operator =(char *s);    // Adquiere el numero de una cadena
NumeroQ operator =(NumeroQ a);
NumeroQ Valor_Absoluto(void); // Devuelve el valor absoluto en OTRO objeto
void Convierte_a_cadena(char *s);   // Copia el numero a una cadena
friend ostream &operator<<(ostream& s,NumeroQ arg);


NumeroQ operator ++(void);
NumeroQ operator --(void);
NumeroQ operator ++(int);
NumeroQ operator --(int);

void operator <<=(ulong l); /* No tiene el sentido propio del desplazamiento,
				pero s el significado del mismo, o sea
				multiplica por la base elevada a 'l'.
				Esto es lo mismo que desplazar el 
				numerador con 'numerador<<=5'    

               a      5   a<<=5
              --- * 10  = -----
               b            b

*/
void operator >>=(ulong l); 
/* 

  a      -5      a       1        a
 --- * 10    = ----- * ----- =  -----
  b                       5     b<<=5
                 b      10       

Esto es lo mismo que desplazar el denominador con 'denominador<<=5'    */

NumeroQ operator +(NumeroQ &a);
NumeroQ operator -(NumeroQ &b);
NumeroQ operator -(void);
NumeroQ operator *(NumeroQ &b);
NumeroQ operator /(NumeroQ &b);

int operator <(NumeroQ &b);
int operator >(NumeroQ &b);
int operator <=(NumeroQ &b);
int operator >=(NumeroQ &b);
int operator ==(NumeroQ &b);
int operator !=(NumeroQ &b);

void Precision(ulong nada);

bool Denom_es_cero(void) {return (Denominador->nPE==0);};

NumeroZ* Num(void) {return Numerador;};
NumeroZ* Den(void) {return Denominador;};

void SetErrorFlag(Tipo_Error e) {Err=e;};
};

#endif
