/**********************************************************************
   numeroz.cc   --  Los objetos de la clase "NumeroZ" pertenecen al conjunto
                    Z de los numeros enteros 

  
    RealNumbers  1.0
    Copyright (C) 1997  Antonio Larrosa Jimenez

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
***********************************************************************/


#include "numeroz.h"
#include "newdebug.h"
#include <stdio.h>
#include "messbox.h"
#include "numfunc.h"

NumeroZ::NumeroZ()
{
Err=No_hay_error;
Base=BASE_N;
Tabla_multiplicar=NULL;
signo=positivo;
maxPE=0;
PE=NULL;
nPE=0;
};

NumeroZ::NumeroZ(NumeroZ &a)
{
Base=a.Base;
Err=a.Err;
Tabla_multiplicar=NULL;
signo=a.signo;
nPE=a.nPE;
maxPE=a.maxPE;
Cifra *psrc;
Cifra *ptgt;
ulong l;
if (maxPE==0)
   PE=NULL;
  else
   {
   PE=(Cifra *)fmalloc(maxPE,IDfmPE);
   psrc=a.PE;
   ptgt=PE;
   l=maxPE;
   while (l>0)
      {
      *ptgt=*psrc;
      ptgt++;
      psrc++;
      l--;
      };
   };
};


NumeroZ::NumeroZ(Numero a)
{
Base=a.Base;
Err=a.Err;
Tabla_multiplicar=NULL;
signo=a.signo;
nPE=a.nPE;
maxPE=a.maxPE;
Cifra *psrc;
Cifra *ptgt;
ulong l;
if (maxPE==0)
   PE=NULL;
  else
   {
   PE=(Cifra *)fmalloc(maxPE,IDfmPE);
   psrc=a.PE;
   ptgt=PE;
   l=maxPE;
   while (l>0)
      {
      *ptgt=*psrc;
      ptgt++;
      psrc++;
      l--;
      };
   };
};

NumeroZ::NumeroZ(long l)
{
Base=BASE_N;
Err=No_hay_error;
Tabla_multiplicar=NULL;
if ((0<=l)&&(l<Base))
   {
   signo=positivo;
   maxPE=1;
   PE=(Cifra *)fmalloc(maxPE,IDfmPE);
   *PE=l;
   nPE=1;
   return;
   };
char s[11];
sprintf(s,"%ld",l);
operator =(s);
};

NumeroZ::NumeroZ(ulong pe,bool inicializa)
{
Base=BASE_N;
Err=No_hay_error;
Tabla_multiplicar=NULL;
signo=positivo;
//maxPE=(pe==0)?(1):(pe); // Siempre debe tener una cifra reservada (por si acaso)
maxPE=pe;
nPE=pe;
if (maxPE>0) PE=(Cifra *)fmalloc(maxPE,IDfmPE);
   else PE=NULL;
if (inicializa==SI)
   {
   Cifra *p=(Cifra *)PE;
   ulong l=pe;
   while (l>0)
       {
       *p=0;
       p++;
       l--;
       };
   };
};

NumeroZ::~NumeroZ()
{
if (PE!=NULL) {ffree(PE);PE=NULL;};
nPE=0;
if (Tabla_multiplicar!=NULL)
    {
    ulong l=1; // La entrada 0 de Tabla_multiplicar siempre esta sin usar.
    while (l<Base)
       {
       if (Tabla_multiplicar[l]!=NULL)
		{
		delete Tabla_multiplicar[l];
		Tabla_multiplicar[l]=NULL;
		};
       l++;
       };
    delete Tabla_multiplicar;
    Tabla_multiplicar=NULL;
    };
};

void NumeroZ::Cambia_signo(void)
{
if (nPE==0) {signo=positivo;return;};
if (signo==positivo) signo=negativo;
    else signo=positivo;
};

NumeroZ NumeroZ::Equal(Cifra c)
{
Nuevo_tamano_PE(1);
nPE=1;
*(PE+maxPE-1)=c;
return (*this);
};

NumeroZ NumeroZ::operator =(long l)
{
if (l==0)
   {
   nPE=0;
   return (*this);
   };
if ((0<l)&&(l<Base))
   {
   signo=positivo;
   Nuevo_tamano_PE(1);
   *(PE+maxPE-1)=(Cifra)l;
   nPE=1;
   return (*this);
   };
char s[11];
sprintf(s,"%ld",l);
operator =(s);
return (*this);
};

NumeroZ NumeroZ::operator =(NumeroZ a)
{
if (maxPE<a.nPE)
   {
   if (PE!=NULL) ffree(PE);
   maxPE=a.maxPE;
   PE=(Cifra *)fmalloc(maxPE,IDfmPE);
   };
signo=a.signo;
nPE=a.nPE;
Base=a.Base;
Err=a.Err;

Cifra *psrc=a.PE+(a.maxPE-a.nPE);
Cifra *ptgt=PE+(maxPE-nPE);
ulong l=nPE;
while (l>0)
   {
   *ptgt=*psrc;
   ptgt++;
   psrc++;
   l--;
   };
// Ya que estamos liados con punteros, podemos mirar si nPE se puede reducir

ptgt=PE+(maxPE-nPE);
while ((nPE>0)&&(*ptgt==0)) {nPE--;ptgt++;};
if (nPE==0) signo=positivo;

return (*this);
};

NumeroZ NumeroZ::operator =(char *s)
{
Err=No_hay_error;
if (Es_un_numero_correcto(s,Base,"-")==NO) return (*this);
if (PE!=NULL) ffree(PE);
if (*s=='-')
   {
   signo=negativo;
   s++;
   }
  else
   if (*s=='+')
      {
      signo=positivo;
      s++;
      }
     else
      signo=positivo;

char *pc=(char *)s;
nPE=0;
ulong l;
while (*pc!=0)
   {
   nPE++;
   pc++;
   };
pc=(char *)s;
bool ok=SI;
if (nPE==0)
   {
   maxPE=0;
   PE=NULL;
   }
  else
   {
   maxPE=nPE;
   PE=(Cifra *)fmalloc(maxPE,IDfmPE);
   l=nPE;
   Cifra *pe=(Cifra *)PE;
   while (l>0)
      {
      *pe=Caracter_a_Cifra(*pc,ok);
      if (ok==NO) {nPE=0;return (*this);};
      pe++;
      pc++;
      l--;
      };
   };
return (*this);
};

ostream &operator<<(ostream& s,NumeroZ arg)
{
  switch (arg.Err)
     {
     case (Division_x_cero) : s << "Division por cero" ; break;
     default : 
             char *pstr=(char *)fmalloc(arg.Longitud()+1,0);
             arg.Convierte_a_cadena(pstr);
             s << pstr;
             ffree(pstr);
     };
  return s;
};

void NumeroZ::Convierte_a_cadena(char *s)
{
//Minimiza_PE();
char *pc=(char *)s;
if (signo==negativo)
   {
   *pc='-';
   pc++;
   };
Cifra *p=(Cifra *)PE;
ulong l;
if (nPE==0)
   {
   *pc='0';
   pc++;
   }
  else   // nPE>0
   {
   l=nPE;
   p+=(maxPE-nPE);
   while (l>0)
      {
      *pc=Cifra_a_Caracter(*p);
      pc++;
      p++;
      l--;
      };
   };
*pc=0;
};

Cifra NumeroZ::Caracter_a_Cifra(char c,bool &ok)
{
ok=SI;
if (((c)>='0')&&((c)<='9')) return (c)-'0';
if (((c)>='A')&&((c)<='Z')) return ((c)-'A')+10;
if (((c)>='a')&&((c)<='z')) return ((c)-'a')+10;
ok=NO;
return CIFRA_IS_NAN;
};

char NumeroZ::Cifra_a_Caracter(Cifra c)
{
if (c<10) return '0'+(c);
if (c==CIFRA_IS_NAN) return '?';
return 'A'+(c-10);
};

NumeroZ NumeroZ::operator +(NumeroZ b)
{
if (signo!=b.signo) // Es una resta
   {
   NumeroZ resultado;
   if (signo==positivo)  //  (+a)+(-b) = (+a)-(+b) = (a-b)
      {
      b.signo=positivo;
      resultado=(*this)-b;
      b.signo=negativo;
      }
     else                //  (-a)+(+b) = (+b)+(-a) = (b-a)
      {
      signo=positivo;
      resultado=b-(*this);
      signo=negativo;
      };
   return resultado;
   };

ulong maxnPE=MAXl(nPE,b.nPE);
NumeroZ c(maxnPE+1,SI);
Cifra carry=0;
ulong l=0;
Cifra ca;
Cifra cb;
Cifra *pc=c.PE+c.maxPE-1;
while (l<maxnPE)
    {
    ca=Cifra_en_posicion_PE(l);
    cb=b.Cifra_en_posicion_PE(l);
    *pc=(ca+cb+carry);
    if (*pc>=Base)
       {
       (*pc)-=Base;
       carry=1;
       }
      else
       carry=0;
    pc--;
    l++;
    };
if (carry==0)
   {
   c.nPE=maxnPE;
   }
  else
   {
   c.nPE=maxnPE+1;
   *pc=carry;
   };
if (signo==negativo) c.signo=negativo;
return c;
};

void NumeroZ::Incrementa(NumeroZ &b)
{
ulong maxnPE=MAXl(nPE,b.nPE);
Nuevo_tamano_PE(maxnPE+1);
Cifra carry=0;
ulong l=0;
Cifra *pc=PE+maxPE-1;
Cifra ca,cb;
while (l<maxnPE)
    {
    ca=Cifra_en_posicion_PE(l);
    cb=b.Cifra_en_posicion_PE(l);
    *pc=(ca+cb+carry);
    if (*pc>=Base)
       {
       (*pc)-=Base;
       carry=1;
       }
      else
       carry=0;
    pc--;
    l++;
    };
if (carry==0)
   {
   nPE=maxnPE;
   }
  else
   {
   nPE=maxnPE+1;
   *pc=carry;
   };
};

Cifra NumeroZ::Cifra_en_posicion_PE(ulong l)
{
if (l>=nPE) return 0;
Cifra *p=PE;
p+=(maxPE-1)-l;
return *p;
};

void NumeroZ::Minimiza_PE()
{
while ((nPE>0)&&(Cifra_en_posicion_PE(nPE-1)==0)) nPE--;
if (nPE==0) signo=positivo;
};

NumeroZ NumeroZ::operator -(NumeroZ b)
{
if (signo!=b.signo) // Es una suma
   {
   NumeroZ resultado;
   if (signo==positivo)  //  (+a)-(-b) = (+a)+(+b) = (a+b)
      {
      b.signo=positivo;
      resultado=(*this)+b;
      b.signo=negativo;
      }
     else                //  (-a)-(+b) = (-a)+(-b) = -(a+b)
      {
      signo=positivo;
      resultado=(*this)+b;
      signo=negativo;
      resultado.signo=negativo;
      };
   return resultado;
   };
ulong maxPE=MAXl(nPE,b.nPE);
NumeroZ c(maxPE+1,SI);
if (signo==positivo)  //Los dos son positivos
   {
   if ((*this)<b)
      {
      c=b-(*this);
      c.Cambia_signo();
      return c;
      };
   }
  else  // Los dos son negativos
   {
      b.signo=positivo;
      signo=positivo;
      c=b-(*this);
      b.signo=negativo;
      signo=negativo;
      return c;
   };

Cifra carry=0;
ulong l=0;
Cifra ca;
Cifra cb;
Cifra *pc=c.PE+maxPE;
while (l<maxPE)
    {
    ca=Cifra_en_posicion_PE(l);
    cb=b.Cifra_en_posicion_PE(l)+carry;
    if (ca<cb)
       {
       *pc=(ca+Base)-cb;
       carry=1;
       }
      else
       {
       *pc=(ca-cb);
       carry=0;
       };
    pc--;
    l++;
    };
*pc=carry;
c.Minimiza_PE();
return c;
};

NumeroZ NumeroZ::operator -(void)
{
NumeroZ Resultado(*this);
Resultado.Cambia_signo();
return Resultado;
};

int NumeroZ::operator <(NumeroZ &b)
{
if (signo!=b.signo)
   {
   if (signo==negativo) return SI;
   return NO;
   };
// Tienen el mismo signo
bool si=SI;
bool no=NO;
if (signo==negativo) // si es negativo el mas grande es el de menos valor
   {
   si=NO;
   no=SI;
   };
if (nPE<b.nPE) return si;
if (nPE>b.nPE) return no;
ulong l=nPE;
Cifra *ca=PE+(maxPE-nPE);
Cifra *cb=b.PE+(b.maxPE-b.nPE);
while (l>0)
    {
//    ca=Cifra_en_posicion_PE(l-1);
//    cb=b.Cifra_en_posicion_PE(l-1);
    if (*ca<*cb) return si;
      else if (*ca>*cb) return no;
    ca++;
    cb++;
    l--;
    };
return no;
};

int NumeroZ::operator >(NumeroZ &b)
{
if (signo!=b.signo)
   {
   if (signo==negativo) return NO;
   return SI;
   };
// Tienen el mismo signo
bool si=SI;
bool no=NO;
if (signo==negativo) // si es negativo el mas grande es el de menos valor
   {
   si=NO;
   no=SI;
   };
if (nPE<b.nPE) return no;
if (nPE>b.nPE) return si;
ulong l=nPE;
Cifra ca,cb;
while (l>0)
    {
    ca=Cifra_en_posicion_PE(l-1);
    cb=b.Cifra_en_posicion_PE(l-1);
    if (ca<cb) return no;
      else if (ca>cb) return si;
    l--;
    };

return no;
};

int NumeroZ::operator >=(NumeroZ &b)
{
if (signo!=b.signo)
   {
   if (signo==negativo) return NO;
   return SI;
   };
// Tienen el mismo signo
bool si=SI;
bool no=NO;
if (signo==negativo) // si es negativo el mas grande es el de menos valor
   {
   si=NO;
   no=SI;
   };
if (nPE<b.nPE) return no;
if (nPE>b.nPE) return si;
ulong l=nPE;
Cifra ca,cb;
while (l>0)
    {
    ca=Cifra_en_posicion_PE(l-1);
    cb=b.Cifra_en_posicion_PE(l-1);
    if (ca<cb) return no;
      else if (ca>cb) return si;
    l--;
    };

return si;
};

int NumeroZ::operator <=(NumeroZ &b)
{
if (signo!=b.signo)
   {
   if (signo==negativo) return SI;
   return NO;
   };
// Tienen el mismo signo
bool si=SI;
bool no=NO;
if (signo==negativo) // si es negativo el mas grande es el de menos valor
   {
   si=NO;
   no=SI;
   };
if (nPE<b.nPE) return si;
if (nPE>b.nPE) return no;
ulong l=nPE;
Cifra ca;
Cifra cb;
while (l>0)
    {
    ca=Cifra_en_posicion_PE(l-1);
    cb=b.Cifra_en_posicion_PE(l-1);
    if (ca<cb) return si;
      else if (ca>cb) return no;
    l--;
    };
return si;
};

int NumeroZ::operator ==(NumeroZ &b)
{
if (signo!=b.signo) return NO;
if (nPE!=b.nPE) return NO;
ulong l=nPE;
Cifra ca,cb;
while (l>0)
    {
    ca=Cifra_en_posicion_PE(l-1);
    cb=b.Cifra_en_posicion_PE(l-1);
    if (ca!=cb) return NO;
    l--;
    };
return SI;
};

int NumeroZ::operator !=(NumeroZ &b)
{
if ((*this)==b) return NO;
return SI;
};

void NumeroZ::Nuevo_tamano_PE(ulong new_num)
{
if (new_num==maxPE) return;
if (new_num==0) return;
//if (new_num==0) new_num=1;  // Por reservar algo siempre
Cifra *pe=(Cifra *)fmalloc(new_num,IDfmPE);
Cifra *psrc=PE;
Cifra *ptgt=pe;
ulong l;
if (new_num>maxPE)
   {
   l=new_num-maxPE;
   while (l>0)
      {
      *ptgt=0;
      ptgt++;
      l--;
      };
   l=maxPE;
   }
  else
   {
   l=new_num;
   psrc+=maxPE-new_num;
   };
while (l>0)
   {
   *ptgt=*psrc;
   psrc++;
   ptgt++;
   l--;
   };
if (PE!=NULL) ffree(PE);
PE=pe;
maxPE=new_num;
};

void NumeroZ::operator <<=(ulong n)
{
if (nPE==0) return;

Cifra *TGT=PE; // Estas tres variables estan para no tener que llamar a
ulong maxTGT=maxPE; // Nuevo_Tamano_PE, asi a la vez reserva la memoria
bool nuevoTGT=NO;   // necesaria y lo copia en su sitio en vez de reservar memoria
                    // copiarla y despues moverla a su sitio
if (maxPE<nPE+n)
   {
   nuevoTGT=SI;
   maxTGT=nPE+n;
   TGT=(Cifra *)fmalloc(maxTGT,IDfmPE);
   };
Cifra *ptgt;
Cifra *psrc;
ptgt=TGT;
ptgt+=(maxTGT-(nPE+n));
psrc=PE;
psrc+=(maxPE-nPE);
ulong l=nPE;
while (l>0)
    {
    *ptgt=*psrc;
    ptgt++;
    psrc++;
    l--;
    };
l=n;
while (l>0)
    {
    *ptgt=0;
    ptgt++;
    l--;
    };
nPE+=n;
if (nuevoTGT==SI)
    {
    if (PE!=NULL) ffree(PE);
    PE=TGT;
    maxPE=maxTGT;
    };
};

void NumeroZ::operator >>=(ulong n)
{
if (n>nPE)
    {
    nPE=0;
    return;
    };
Cifra *ptgt;
Cifra *psrc;
ptgt=PE+(maxPE-1);
psrc=PE+(maxPE-n-1);
ulong l=nPE-n;
while (l>0)
    {
    *ptgt=*psrc;
    ptgt--;
    psrc--;
    l--;
    };
nPE-=n;
};

Cifra NumeroZ::CifraxCifra(Cifra a,Cifra b,Cifra &Carry)
{
Double_Cifra tmp;
tmp=(a*b)+Carry;
Double_Cifra Carrytmp=0;
while ((Carrytmp+1)*Base<=tmp)
   {
   Carrytmp++;
   };
Carry=Carrytmp;
tmp-=Carrytmp*Base;
return tmp;
};

void NumeroZ::Multiplica_por_cifra(Cifra c,NumeroZ *Resultado)
{
if (c==0)
   {
   Resultado->Equal(0);
   return;
   }
if (c==1)
   {
   (*Resultado)=(*this);
   return;
   };

Resultado->Nuevo_tamano_PE(nPE+1);
ulong l=0;
Cifra *pe=PE+maxPE-1;
Cifra *ptgt=Resultado->PE+Resultado->maxPE-1;
Cifra carry=0;
while (l<nPE)
   {
   *ptgt=CifraxCifra(c,*pe,carry);
   ptgt--;
   pe--;
   l++;
   };
if (carry==0)
   {
   Resultado->nPE=nPE;
   }
  else
   {
   *ptgt=carry;
   Resultado->nPE=nPE+1;
   };
};

NumeroZ NumeroZ::Multiplica_normal(NumeroZ &b)
{
Cifra *pB=b.PE+b.maxPE-1;
Inicializa_tabla_multiplicar();
ulong l=0;
NumeroZ Temporal;//(maxPE+b.maxPE,0);
NumeroZ Resultado;
short i; // Aqui uso un short en vez de un tipo Cifra, porque para acceder
	//  a un array el tipo indice no puede ser del tipo Cifra
while (l<b.nPE)
   {
   i=*pB;
   if (Tabla_multiplicar[i]==NULL)
       {
       Tabla_multiplicar[i]=new NumeroZ(nPE+1,NO);
       Multiplica_por_cifra(*pB,Tabla_multiplicar[i]);
       };
   Temporal=*Tabla_multiplicar[i];
   Temporal<<=l;
   Resultado.Incrementa(Temporal);
   pB--;
   l++;
   };
return Resultado;
};

NumeroZ NumeroZ::operator *(NumeroZ &b)
{
if ((nPE==0)||(b.nPE==0))
    {
    NumeroZ Cero;
    return Cero;
    };
NumeroZ Resultado(Multiplica_normal(b));
if (b.signo!=signo)
   Resultado.signo=negativo;
return Resultado;
};

NumeroZ NumeroZ::Factorial()
{
NumeroZ Resultado;
if ((*this)<Resultado) return Resultado;
Resultado++;

NumeroZ Tmp(nPE,SI);  /* Deberia de poner 'NumeroZ Tmp(1);' pero de esta 
			forma tiene allocateada la memoria desde el principio */
Tmp.nPE=0;
++Tmp;
NumeroZ Uno(1);
while (Tmp<(*this))
   {
   Resultado=Resultado.Multiplica_normal(Tmp);
		/* Se pueden hacer suposiciones respecto a los signos
		que es la unica diferencia entre usar el operador *
		y la funcion Multiplica_normal */
   ++Tmp;
   };
Resultado=Resultado*Tmp;
return Resultado;
};

NumeroZ NumeroZ::operator ++(void)
{
if (nPE==0)
   {
   Nuevo_tamano_PE(1);
   *(PE+maxPE-1)=1;
   nPE=1;
   signo=positivo;
   return (*this);
   };
if (signo==negativo)
   {
   if ((nPE==1)&&(*(PE+maxPE-1)==1)) // (-1)++  que es 0, lo hacemos cableado
      {
      nPE=0;
      signo=positivo;
      return (*this);
      };
   signo=positivo;
   operator--();
   signo=negativo;
   return (*this);
   };
Cifra *pe=PE;
pe+=maxPE-1;
(*pe)++;
ulong l=nPE;
while ((*pe>=Base)&&(l>1))
   {
   *pe-=Base; // Y me llevo un carry de 1
   pe--;
   l--;
   (*pe)++;
   };
if ((l==1)&&(*pe>=Base))  // Hay que ampliar en una cifra
   {
   Nuevo_tamano_PE(nPE+1);
   *PE=1; // El carry es de 1 seguro
   (*(PE+1))-=Base;
   nPE++;
   };
return (*this);
};

NumeroZ NumeroZ::operator --(void)
{
if (nPE==0) // Si es 0 al operarlo queda -1 con lo que se puede hacer cableado
   {
   Nuevo_tamano_PE(1);
   *(PE+maxPE-1)=1;
   nPE=1;
   signo=negativo;
   return (*this);
   };
if (signo==negativo)
   {
   signo=positivo;
   operator++();
   signo=negativo;
   return (*this);
   };

Cifra *pe=PE;
pe+=maxPE-1;
bool fin=NO;
// En el siguiente bucle no hace falta tener en cuenta a nPE porque se puede
//  suponer que las variables son consistentes y entonces se comprueba que
//  SIEMPRE acaba, excepto en el caso del 0 pero este caso ya se comprob al
//  principio de la funcin
while (!fin)
   {
   if (*pe!=0)
      {
      (*pe)--;
      fin=SI;
      }
     else
      {
      (*pe)=Base-1;
      };
   pe--;
   };
return (*this);
};

NumeroZ NumeroZ::operator ++(int)
{
NumeroZ Antes_de_cambios(*this);
operator++();
return Antes_de_cambios;
};

NumeroZ NumeroZ::operator --(int)
{
NumeroZ Antes_de_cambios(*this);
operator--();
return Antes_de_cambios;
};

void NumeroZ::Anade_a_la_derecha_en_PE(Cifra c)
{
if ((nPE==0)&&(c==0)) return; // Esto ocurre si ya es 0 y se le va a aadir un 0

Nuevo_tamano_PE(nPE+1);
ulong l=nPE;
Cifra *psrc=PE+(maxPE-nPE);
Cifra *ptgt=PE+(maxPE-nPE-1);
while (l>0)
   {
   *ptgt=*psrc;
   ptgt++;
   psrc++;
   l--;
   };
*ptgt=c;
nPE++;
};

void NumeroZ::Inicializa_tabla_multiplicar(void)
{
ulong l=0;
if (Tabla_multiplicar!=NULL)
    {
    if ((*Tabla_multiplicar[1])==(*this)) return;
    while (l<Base)
       {
       if (Tabla_multiplicar[l]!=NULL)
           {
           delete Tabla_multiplicar[l];
           Tabla_multiplicar[l]=NULL;
           };
       l++;
       };
//    delete Tabla_multiplicar;
    }
   else
    {
    Tabla_multiplicar=new (NumeroZ*)[Base];
    while (l<Base)
       {
       Tabla_multiplicar[l]=NULL;
       l++;
       };
    };
//   Tabla_multiplicar[0]=new Numero;
Tabla_multiplicar[1]=new NumeroZ(*this);// Aqui se copia a si mismo para poder
       // reusar la tabla si se hacen varias multiplicaciones seguidas con el mismo
       // numero. Asi sabe despues si se puede reusar o no (con la comparacion del
       // principio de esta funcion)
};

Cifra NumeroZ::Divide_N_entre_N(NumeroZ *r,NumeroZ &divisor)
{
if (*r<divisor) return 0;
if ((*r)>=(*divisor.Tabla_multiplicar[Base-1]))
   {
   (*r)=(*r)-(*divisor.Tabla_multiplicar[Base-1]);
   return Base-1;
   };

// Esta funcion hace una "Busqueda binaria" del resultado, y va guardando
// algunos resultados en un array por si en una llamada posterior a esta
// funcion lo necesita, poder ir mas rapido.
Cifra a=1;     // Cota inferior
Cifra b=Base-1;  // Cota superior
short i;
while (b-a>1)
   {
   i=(a+b)/2;
   if (divisor.Tabla_multiplicar[i]==NULL)
      {
      divisor.Tabla_multiplicar[i]=new NumeroZ(divisor.nPE+1,NO);
      divisor.Multiplica_por_cifra(i,divisor.Tabla_multiplicar[i]);
      };
   if ((*divisor.Tabla_multiplicar[i])<(*r))
       {
       a=i;
       }
      else
       if ((*divisor.Tabla_multiplicar[i])>(*r))
           {
           b=i;
           }
          else // Es igual
           {
           r->nPE=0;
           return i;
           };
   };
(*r)=(*r)-(*divisor.Tabla_multiplicar[a]);
return a;
};

NumeroZ NumeroZ::operator /(NumeroZ &divisor)
{
NumeroZ Cociente;
if (divisor.nPE==0) {Cociente.Err=Division_x_cero;return Cociente;}; 
sign signodivisor=divisor.signo;
divisor.signo=positivo;

NumeroZ Temporal;

Division_da_cociente_y_resto(divisor,Cociente,Temporal);

if (signo!=signodivisor) Cociente.signo=negativo;
divisor.signo=signodivisor;

return Cociente;
};

NumeroZ NumeroZ::operator %(NumeroZ &divisor)
{ // Esta es la funcion modulo, particiona todo Z de la siguiente forma :
//  si divisor es 3 :
//  Z :  ...  -5  -4  -3  -2  -1  0  1  2  3  4  5  ...
//       ...   1   2   0   1   2  0  1  2  0  1  2  ...
//  si divisor es -3 :
//  Z :  ...  -5  -4  -3  -2  -1  0  1  2  3  4  5  ...
//       ...   2   1   0   2   1  0  2  1  0  2  1  ...

sign signodivisor=divisor.signo;
divisor.signo=positivo;
/*divisor.Inicializa_tabla_multiplicar();
// Las tres siguientes lineas terminan de inicializar la tabla de multiplicar
//  ya que al dividir es necesario que est preparado la multiplicacion por
//  la ultima cifra posible de la Base
short i=Base-1;
if (divisor.Tabla_multiplicar[i]==NULL) divisor.Tabla_multiplicar[i]=new NumeroZ(nPE+1,NO);
divisor.Multiplica_por_cifra(i,divisor.Tabla_multiplicar[i]);

NumeroZ Cociente;
NumeroZ Temporal;
ulong l=nPE;
Cifra *p=PE+(maxPE-nPE);
Cifra c;
// Empieza cogiendo de la parte entera ...
while (l>0)
   {
   Temporal.Anade_a_la_derecha_en_PE(*p);
   c=Divide_N_entre_N(&Temporal,divisor);
   Cociente.Anade_a_la_derecha_en_PE(c);
   p++;
   l--;
   };
*/
NumeroZ Cociente;
NumeroZ Temporal;

Division_da_cociente_y_resto(divisor,Cociente,Temporal);

if (signo!=signodivisor)
   {
   Temporal=divisor-Temporal;
   }
divisor.signo=signodivisor;

return Temporal;
};

void NumeroZ::Division_da_cociente_y_resto(NumeroZ &divisor,NumeroZ &Cociente,NumeroZ &Resto)
{
divisor.Inicializa_tabla_multiplicar();
// Las tres siguientes lineas terminan de inicializar la tabla de multiplicar
//  ya que al dividir es necesario que est preparado la multiplicacion por
//  la ultima cifra posible de la Base
short i=Base-1;
if (divisor.Tabla_multiplicar[i]==NULL) divisor.Tabla_multiplicar[i]=new NumeroZ(nPE+1,NO);
divisor.Multiplica_por_cifra(i,divisor.Tabla_multiplicar[i]);

Cociente.nPE=0;
Resto.nPE=0;
ulong l=nPE;
Cifra *p=PE+(maxPE-nPE);
Cifra c;
// Empieza cogiendo de la parte entera ...
while (l>0)
   {
   Resto.Anade_a_la_derecha_en_PE(*p);
   c=Divide_N_entre_N(&Resto,divisor);
   Cociente.Anade_a_la_derecha_en_PE(c);
   p++;
   l--;
   };
};

NumeroZ NumeroZ::Valor_Absoluto(void)
{
NumeroZ resultado(*this);
resultado.signo=positivo;
return resultado;
};

ulong NumeroZ::Longitud(void)
{
  return nPE+((signo==positivo)?(0):(1));
};

void NumeroZ::Precision(ulong)
{
};

NumeroZ NumeroZ::MCD(NumeroZ &b)
{
NumeroZ A(*this);
NumeroZ B(b);
A.signo=positivo;
B.signo=positivo;
NumeroZ tmp;
while ((B.nPE!=0)&&(A.nPE!=0))
	{
//	cout << A << " , " << B << endl;
	if (A>B) A=A%B;
	   else
	    {
	    tmp=A;
	    A=B%A;
	    B=tmp;
	    };
	};
//cout << A << " , " << B << endl;
if (A.nPE==0) return B;
return A;
};
