/**********************************************************************
    numeroz.h
  
    RealNumbers  1.0
    Copyright (C) 1997  Antonio Larrosa Jimenez

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
***********************************************************************/
#ifndef _NUMEROZ_H
#define _NUMEROZ_H

#include "dattypes.h"
#include "numero.h"
#include "numconst.h"
#include "errores.h"

class NumeroZ
{
friend Numero;
friend NumeroQ;
private:
Tipo_Error Err;

sign signo; // El signo del numero

Cifra Base;

Cifra *PE; // Parte entera

ulong nPE; // numero de cifras de la parte entera

ulong maxPE; // El numero de cifras RESERVADAS en la parte entera

Cifra Caracter_a_Cifra(char c,bool &ok); // Convierte de ASCII a byte y devuelve ok=NO si no es una cifra permitida (por ejemplo un signo $)
char Cifra_a_Caracter(Cifra c); // Convierte de ASCII a byte

void Nuevo_tamano_PE(ulong new_units);

void Incrementa(NumeroZ &b);      // Es parecido al significado de += pero
                                  //  incrementa sin tener en cuenta el signo
                                  //  (incrementa el valor absoluto de b)
NumeroZ Multiplica_normal(NumeroZ &b);// Multiplica (*this) por b y devuelve
                                      // el resultado, pero sin tener en cuenta
                                      // los signos
Cifra CifraxCifra(Cifra a,Cifra b,Cifra &Carry);
void Multiplica_por_cifra(Cifra c,NumeroZ *Resultado);

NumeroZ **Tabla_multiplicar;
// Al inicializarlo se puede usar como un array que tiene Base elementos
//  es decir, como si la definicion fuera Numero *Tabla_multiplicar[Base];
//  Al hacer una division a mano, multiplicamos el divisor por varios numeros
//  para obtener un numero que aadir al cociente, si el numero es grande,
//  se repiten las multiplicaciones, por qu no guardarlas y as no
//  hay que repetir la misma operacion?. Para eso sirve esta variable
void Inicializa_tabla_multiplicar(void);
Cifra Divide_N_entre_N(NumeroZ *r,NumeroZ &divisor);

void Anade_a_la_derecha_en_PE(Cifra c);
     //Basicamente, hacer a.Anade_a_la_derecha(c); es como hacer :
                                  //   a<<=1;
                                  //   *(PE+maxPE-1)=c;

public:
NumeroZ();
NumeroZ(long l);
NumeroZ(NumeroZ &a);
NumeroZ(Numero a);
NumeroZ(ulong pe,bool inicializa); // Tamao de PE y si debe rellenarlo con 0
~NumeroZ();

ulong NPE(void) {return nPE;};
Cifra Cifra_en_posicion_PE(ulong l); // Desde 0 (LSB) al infinito (MSB)

sign Signo(void) {return signo;};
NumeroZ operator =(long l);
NumeroZ Equal(Cifra c);
NumeroZ operator =(char *s);    // Adquiere el numero de una cadena
NumeroZ operator =(NumeroZ a);
void Convierte_a_cadena(char *s);   // Copia el numero a una cadena
friend ostream &operator<<(ostream& s,NumeroZ arg);

void Cambia_signo(void);
NumeroZ Valor_Absoluto(void);

NumeroZ operator ++(void);
NumeroZ operator --(void);
NumeroZ operator ++(int);
NumeroZ operator --(int);

void operator <<=(ulong l); /* Esta funcion se puede usar para multiplicar
                             por la base, por ejemplo si la base es 10 y
                             se quiere multiplicar por 100, pues se pone
                                      numero<<=2;
                             que es la misma sentencia que se pondria si la
                             base fuera 2 y se quisiera multiplicar por 4 */
void operator >>=(ulong l); /* Lo mismo que arriba, simplemente desplaza a la
                             derecha o divide */

NumeroZ operator +(NumeroZ a);
NumeroZ operator -(NumeroZ b);
NumeroZ operator -(void);
NumeroZ operator *(NumeroZ &b);
NumeroZ operator /(NumeroZ &b);
NumeroZ operator %(NumeroZ &b);

NumeroZ Factorial(void);
NumeroZ MCD(NumeroZ &b);  // Calcula el maximo comun divisor entre *this y b
void Division_da_cociente_y_resto(NumeroZ &divisor,NumeroZ &Cociente,NumeroZ &Resto);

int operator <(NumeroZ &b);
int operator >(NumeroZ &b);
int operator <=(NumeroZ &b);
int operator >=(NumeroZ &b);
int operator ==(NumeroZ &b);
int operator !=(NumeroZ &b);

void Minimiza_PE();


ulong Longitud(void);
void Precision(ulong nada);
void SetErrorFlag(Tipo_Error e) {Err=e;};
};

#endif
